/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// Taken from DX8 sample

#include "stdafx.h"
#include <dinput.h>
#include <atlbase.h>		// CComBSTR

#include "ConfigHandler.h"
#include "DInputHandler.h"
#include "DBGConsole.h"


static LPDIRECTINPUT8       g_pDI       = NULL;

#define ANALOGUE_STICK_RANGE		80

static BOOL CALLBACK EnumJoysticksCallback( const DIDEVICEINSTANCE* pdidInstance, VOID* pContext );
static BOOL CALLBACK EnumAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, VOID* pContext );

static BOOL CALLBACK EnumAllAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, LPVOID pContext );

static void Input_ReadConfig();
static void Input_WriteConfig();
static HRESULT Input_CreateDevices();
static void Input_FreeDevices();

static HWND	s_hWnd = NULL;
static std::vector< GUID > g_DeviceGuids;
static DWORD s_dwNumDevices = 0;
static LPDIRECTINPUTDEVICE8 * g_pDeviceInstances;

// As we may be called by several threads, we use this
// critical section to ensure that only one thread can 
// access the above arrays at once.
static CCritSect	s_InputSemaphore;

// This basically says if the n64 control is a button or a 
// axis, and gives the button bits if it is a button
N64Button g_N64Buttons[NUM_INPUT_IDS] =
{
	{ TEXT("Analogue Stick X"), 0, TRUE },		// Analogue x-axis
	{ TEXT("Analogue Stick Y"), 0, TRUE },		// Analogue y-axis
	{ TEXT("DPad Up"), U_JPAD, FALSE },			// Up Digital
	{ TEXT("DPad Left"), L_JPAD, FALSE },		// Left Digital
	{ TEXT("DPad Right"), R_JPAD, FALSE },		// Right Digital
	{ TEXT("DPad Down"), D_JPAD, FALSE },		// Down Digital
	{ TEXT("A Button"), A_BUTTON, FALSE },		// A
	{ TEXT("B Button"), B_BUTTON, FALSE },		// B
	{ TEXT("Z Trigger"), Z_TRIG, FALSE },		// Z
	{ TEXT("Left Pan"), L_TRIG, FALSE },		// L Trig
	{ TEXT("Right Pan"), R_TRIG, FALSE },		// R Trig
	{ TEXT("C Up"), U_CBUTTONS, FALSE },		// Up C
	{ TEXT("C Left"), L_CBUTTONS, FALSE },		// Left C
	{ TEXT("C Right"), R_CBUTTONS, FALSE },		// Right C
	{ TEXT("C Down"), D_CBUTTONS, FALSE },		// Down C
	{ TEXT("Start"), START_BUTTON, FALSE },		// Start
};




// In the default assignment, all buttons are assigned to the keyboard
static const ButtonAssignment s_DefaultButtons[NUM_INPUT_IDS] =
{
	// 0 is for keyboard, by default input device 0
	{ 0, FALSE, DIK_LEFT,	DIK_RIGHT },
	{ 0, FALSE, DIK_UP,		DIK_DOWN },
	{ 0, FALSE, DIK_T,		0 },			// Up Digital
	{ 0, FALSE, DIK_F,		0 },			// Left Digital
	{ 0, FALSE, DIK_G,		0 },			// Right Digital
	{ 0, FALSE, DIK_V,		0 },			// Down Digital
	{ 0, FALSE, DIK_A,		0 },			// A
	{ 0, FALSE, DIK_S,		0 },			// B
	{ 0, FALSE, DIK_X,		0 },			// Z
	{ 0, FALSE, DIK_LBRACKET, 0 },			// L Trig
	{ 0, FALSE, DIK_RBRACKET, 0 },			// R Trig
	{ 0, FALSE, DIK_I,		0 },			// Up C
	{ 0, FALSE, DIK_J,		0 },			// Left C
	{ 0, FALSE, DIK_K,		0 },			// Right C
	{ 0, FALSE, DIK_M,		0 },			// Down C
	{ 0, FALSE, DIK_RETURN, 0, },			// Start
};

// This is set up either by reading the config, or by rescannign the devices
ButtonAssignment g_CurrentButtons[NUM_INPUT_IDS];





//-----------------------------------------------------------------------------
// Name: Input_Initialise()
// Desc: Initialize the DirectInput variables.
//-----------------------------------------------------------------------------
HRESULT Input_Initialise( HWND hWnd )
{
    HRESULT hr;

	s_hWnd = hWnd;

    // Register with the DirectInput subsystem and get a pointer
    // to a IDirectInput interface we can use.
    // Create a DInput object
    if( FAILED( hr = DirectInput8Create( g_hInstance, DIRECTINPUT_VERSION, 
                                         IID_IDirectInput8, (VOID**)&g_pDI, NULL ) ) )
        return hr;

	Input_ReadConfig();
	if (s_dwNumDevices == 0)
	{
		Input_RescanDevices();
	}

	// At this point the devices will have been created, and 
	// the initial button assigments will have been created

    return S_OK;
}

BOOL CALLBACK EnumAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, LPVOID pContext )
{
    LPDIRECTINPUTDEVICE8 pDev = (LPDIRECTINPUTDEVICE8)pContext;

    DIPROPRANGE diprg; 
    diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
    diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
    diprg.diph.dwHow        = DIPH_BYID; 
    diprg.diph.dwObj        = pdidoi->dwType; // Specify the enumerated axis

	// We set the range to -80 to +80 here
    diprg.lMin              = -ANALOGUE_STICK_RANGE; 
    diprg.lMax              = +ANALOGUE_STICK_RANGE; 
    
	// Set the range for the axis
	if( FAILED( pDev->SetProperty( DIPROP_RANGE, &diprg.diph ) ) )
		return DIENUM_STOP;

    return DIENUM_CONTINUE;
}



//-----------------------------------------------------------------------------
// Name: FreeDirectInput()
// Desc: Initialize the DirectInput variables.
//-----------------------------------------------------------------------------
void Input_Finalise()
{
	Input_WriteConfig();
	Input_FreeDevices();

    SAFE_RELEASE( g_pDI );

	s_InputSemaphore.Unlock();

}


static void Input_ReadConfig()
{
	HRESULT hr;
	LONG i;
	LONG n;
	TCHAR szBuffer[200+1];
	ConfigHandler * pConfig = new ConfigHandler("Input");

	s_InputSemaphore.Lock();

	// Free any devices that we have acquired
	Input_FreeDevices();

	// Copy default values to current
	memcpy(g_CurrentButtons, s_DefaultButtons, sizeof(s_DefaultButtons));
	
	if (pConfig != NULL)
	{
		i = 0;
		do
		{
			pConfig->ReadString("Device", szBuffer, sizeof(szBuffer), "", i);

			if (lstrlen(szBuffer) > 0)
			{
				GUID guid;
				CComBSTR bstrGuid(szBuffer);

				hr = IIDFromString(bstrGuid, &guid);
				if (SUCCEEDED(hr))
					g_DeviceGuids.push_back(guid);
			}

			i++;
		}
		while (lstrlen(szBuffer) > 0);

		/*hr = */Input_CreateDevices();

		// Now read in controls
		for (i = 0; i < NUM_INPUT_IDS; i++)
		{
			ButtonAssignment ba;

			pConfig->ReadString(g_N64Buttons[i].szName, szBuffer, sizeof(szBuffer), "");

			n = sscanf(szBuffer, "%d %d %d %d",
				&ba.dwDevice, 
				&ba.bIsAxis,
				&ba.dwOfsA,
				&ba.dwOfsB);

			// Only assign if fields are valid
			if (n == 4 && ba.dwDevice < s_dwNumDevices)
			{
				g_CurrentButtons[i] = ba;
			}
		}
	}
	s_InputSemaphore.Unlock();
}

void Input_WriteConfig()
{
	HRESULT hr;
	LONG i;
	TCHAR szBuffer[200+1];

	ConfigHandler * pConfig = new ConfigHandler("Input");

	s_InputSemaphore.Lock();
	if (pConfig != NULL)
	{
		// Write out device data
		for (i = 0; i < s_dwNumDevices; i++)
		{
			LPOLESTR szString;

			hr = StringFromIID(g_DeviceGuids[i], &szString);
			if (SUCCEEDED(hr))
			{
				IMalloc * pMalloc = NULL;

				wsprintf(szBuffer, "%S", szString);
				pConfig->WriteString("Device", szBuffer, i);

				// Free the buffer
				hr = CoGetMalloc(1, &pMalloc);
				if (SUCCEEDED(hr))
				{
					pMalloc->Free(szString);
					pMalloc->Release();
				}

			}
		}


		// Write out button assignments
		for (i = 0; i < NUM_INPUT_IDS; i++)
		{
			wsprintf(szBuffer, "%d %d %d %d",
				g_CurrentButtons[i].dwDevice, 
				g_CurrentButtons[i].bIsAxis,
				g_CurrentButtons[i].dwOfsA,
				g_CurrentButtons[i].dwOfsB);
			pConfig->WriteString(g_N64Buttons[i].szName, szBuffer);
		}

		delete pConfig;
	}
	s_InputSemaphore.Unlock();


}

// Rescan for any attached devices. This resets the current config
void Input_RescanDevices()
{
	s_InputSemaphore.Lock();

	// Release any other devices we have acquired
	Input_FreeDevices();
	
	// Always include the KeyBoard!
	g_DeviceGuids.push_back(GUID_SysKeyboard);

	// Look for any attached joysticks - this basically
	// pushes the guids of any devices we disconver on
	// to g_DeviceGuids
	/*hr = */g_pDI->EnumDevices( DI8DEVCLASS_GAMECTRL, 
								 EnumJoysticksCallback,
								 NULL, DIEDFL_ATTACHEDONLY );
	
	/*hr = */Input_CreateDevices();

	// Copy default values to current
	memcpy(g_CurrentButtons, s_DefaultButtons, sizeof(s_DefaultButtons));

	s_InputSemaphore.Unlock();
}

BOOL CALLBACK EnumJoysticksCallback( const DIDEVICEINSTANCE* pdidInstance, LPVOID pContext )
{
	// Maybe do some checks on the type here

	// For now just push the guid onto our list
	g_DeviceGuids.push_back(pdidInstance->guidInstance);

    return DIENUM_CONTINUE;
}


void Input_FreeDevices()
{
	LONG i;

	s_InputSemaphore.Lock();
	if (g_pDeviceInstances != NULL)
	{
		for (i = 0; i < s_dwNumDevices; i++)
		{
			if (g_pDeviceInstances[i] != NULL)
			{
				g_pDeviceInstances[i]->Unacquire();
				g_pDeviceInstances[i]->Release();
			}
		}
		delete [] g_pDeviceInstances;
	}	
	g_DeviceGuids.clear();
	s_dwNumDevices = 0;
	s_InputSemaphore.Unlock();
}

// IMPORTANT! Must be called from within the critical section!
HRESULT Input_CreateDevices()
{
	LONG i;
	HRESULT hr;

	// Size g_DeviceInstances to the same size
	g_pDeviceInstances = new LPDIRECTINPUTDEVICE8[g_DeviceGuids.size()];
	if (g_pDeviceInstances == NULL)
		return E_OUTOFMEMORY;

	s_dwNumDevices = g_DeviceGuids.size();

	// Clear array
	ZeroMemory(g_pDeviceInstances, sizeof(LPDIRECTINPUTDEVICE8) * s_dwNumDevices);

	for (i = 0; i < s_dwNumDevices; i++)
	{
		LPDIRECTINPUTDEVICE8 pDevice = NULL;

		// Device 1 is the first joystick
	    hr = g_pDI->CreateDevice( g_DeviceGuids[i], &pDevice, NULL );

		if SUCCEEDED(hr)
		{
			if (i == 0)
			{
				hr = pDevice->SetDataFormat( &c_dfDIKeyboard );
				if (FAILED(hr))
				{
					pDevice->Release();
					continue;		// Try the next device
				}

			}
			else
			{
				// Set the data format to "simple joystick" - a predefined data format 
				//
				// A data format specifies which controls on a device we are interested in,
				// and how they should be reported. This tells DInput that we will be
				// passing a DIJOYSTATE2 structure to IDirectInputDevice::GetDeviceState().
				hr = pDevice->SetDataFormat( &c_dfDIJoystick2 );
				if (FAILED(hr))
				{
					pDevice->Release();
					continue;		// Try the next device
				}
			}

			// Set the cooperative level to let DInput know how this device should
			// interact with the system and with other DInput applications.
			hr = pDevice->SetCooperativeLevel( s_hWnd, DISCL_EXCLUSIVE | 
													   DISCL_FOREGROUND );
			if (FAILED(hr))
			{
				pDevice->Release();
				continue;
			}

			// Determine how many axis the joystick has (so we don't error out setting
			// properties for unavailable axis)
			//g_diDevCaps.dwSize = sizeof(DIDEVCAPS);
			//if ( FAILED( hr = pDevice->GetCapabilities(&g_diDevCaps) ) )
			//	return hr;

			// Enumerate the axes of the joystick and set the range of each axis. Note:
			// we could just use the defaults, but we're just trying to show an example
			// of enumerating device objects (axes, buttons, etc.).
			hr = pDevice->EnumObjects( EnumAxesCallback, 
									(VOID*)pDevice, DIDFT_AXIS );

			if (FAILED(hr))
			{
				pDevice->Release();
				continue;
			}
			g_pDeviceInstances[i] = pDevice;
		}

	}

	return S_OK;
}


// We need this struct to pass in both control list and the device instance info
typedef struct
{
	ControlList * pcl;
	DWORD dwDevice;

	TCHAR szDeviceName[MAX_PATH+1];

} EnumObjectInfo;


HRESULT Input_GenerateAllObjects(ControlList * pCL)
{
	LONG b;
	LONG i;
	HRESULT hr;
	DIDEVICEINSTANCE di;
	DIDEVCAPS diDevCaps;

	if (pCL == NULL)
		return E_POINTER;

	pCL->controls.clear();


	s_InputSemaphore.Lock();
	for (i = 0; i < s_dwNumDevices; i++)
	{
		// Skip any devices we couldn't create
		if (g_pDeviceInstances[i] == NULL)
			continue;

		// Get the device info so that we can pass it into the enumeration function
		di.dwSize = sizeof(di);
		hr = g_pDeviceInstances[i]->GetDeviceInfo(&di);
		if (FAILED(hr))
			continue;

		if (i == 0)
		{
			EnumObjectInfo eoi;
			// For the keyboard, enumerate all the buttons so that we
			// can get the diplay names
			lstrcpyn(eoi.szDeviceName, di.tszInstanceName, sizeof(eoi.szDeviceName));

			eoi.pcl = pCL;
			eoi.dwDevice = i;

			/*hr = */g_pDeviceInstances[i]->EnumObjects( EnumAllAxesCallback, (LPVOID)&eoi, DIDFT_AXIS|DIDFT_BUTTON);

		}
		else
		{
			ControlID cid;

			diDevCaps.dwSize = sizeof(DIDEVCAPS);
			hr = g_pDeviceInstances[i]->GetCapabilities(&diDevCaps);
			if (FAILED(hr))
				continue;

			cid.dwDevice = i;

			if (diDevCaps.dwAxes >= 1)
			{
				cid.dwOfs = DIJOFS_X;
				wsprintf(cid.szName, "%s.X Axis", di.tszInstanceName);
				cid.bIsAxis = TRUE;
				pCL->controls.push_back(cid);
			}
			if (diDevCaps.dwAxes >= 2)
			{
				cid.dwOfs = DIJOFS_Y;
				wsprintf(cid.szName, "%s.Y Axis", di.tszInstanceName);
				cid.bIsAxis = TRUE;
				pCL->controls.push_back(cid);
			}
			// Ignore Z axis

			for (b = 0; b < diDevCaps.dwButtons; b++)
			{
				cid.dwOfs = DIJOFS_BUTTON(b);
				cid.bIsAxis = FALSE;
				wsprintf(cid.szName, "%s.Button %d", di.tszInstanceName, b, cid.dwOfs);
				pCL->controls.push_back(cid);
				// Add button
			}
		}



	}
	s_InputSemaphore.Unlock();

	return S_OK;
}

static BOOL Input_IsObjectOnExcludeList( DWORD dwOfs )
{
	if (dwOfs == DIK_PREVTRACK  ||
	    dwOfs == DIK_NEXTTRACK  ||
	    dwOfs == DIK_MUTE       ||
	    dwOfs == DIK_CALCULATOR ||
	    dwOfs == DIK_PLAYPAUSE  ||
	    dwOfs == DIK_MEDIASTOP  ||
	    dwOfs == DIK_VOLUMEDOWN ||
	    dwOfs == DIK_VOLUMEUP   ||
	    dwOfs == DIK_WEBHOME    ||
	    dwOfs == DIK_SLEEP      ||
	    dwOfs == DIK_WEBSEARCH  ||
	    dwOfs == DIK_WEBFAVORITES ||
	    dwOfs == DIK_WEBREFRESH ||
	    dwOfs == DIK_WEBSTOP    ||
	    dwOfs == DIK_WEBFORWARD ||
	    dwOfs == DIK_WEBBACK    ||
	    dwOfs == DIK_MYCOMPUTER ||
	    dwOfs == DIK_MAIL       ||
		dwOfs == DIK_WAKE		||
		dwOfs == DIK_POWER		||
	    dwOfs == DIK_MEDIASELECT)
		return TRUE;

	return FALSE;
}

BOOL CALLBACK EnumAllAxesCallback( const DIDEVICEOBJECTINSTANCE* pdidoi, LPVOID pContext )
{
	ControlID cid;
	EnumObjectInfo * peoi;


	peoi = (EnumObjectInfo *)pContext;

	if (pdidoi->guidType == GUID_Key)
	{
		if (Input_IsObjectOnExcludeList(pdidoi->dwOfs))
			return DIENUM_CONTINUE;

	}

	cid.dwDevice = peoi->dwDevice;
	cid.dwOfs = pdidoi->dwOfs;

	wsprintf(cid.szName, "%s.%s", peoi->szDeviceName, pdidoi->tszName);


    // Set the UI to reflect what axes the joystick supports
	if (pdidoi->guidType == GUID_XAxis ||
		pdidoi->guidType == GUID_YAxis)
	{
		cid.bIsAxis = TRUE;
		peoi->pcl->controls.push_back(cid);
	}
	else if (pdidoi->guidType == GUID_Key ||
			 pdidoi->guidType == GUID_Button)
	{
		cid.bIsAxis = FALSE;
		peoi->pcl->controls.push_back(cid);
	}
	else
	{
		// Ignore
	}

    return DIENUM_CONTINUE;
}

// Unaquire the input focus, if we have it
void Input_Unaquire()
{
	LONG i;

	s_InputSemaphore.Lock();

	// Release any other devices we have acquired
	if (g_pDeviceInstances != NULL)
	{
		for (i = 0; i < s_dwNumDevices; i++)
		{
			if (g_pDeviceInstances[i] != NULL)
				g_pDeviceInstances[i]->Unacquire();
		}
	}	

	s_InputSemaphore.Unlock();
}


HRESULT Input_GetState( OSContPad * pPad )
{
	LONG i;
	LONG c;
    HRESULT     hr;
	BYTE    diks[256];   // DirectInput keyboard state buffer 
	DIJOYSTATE2 js;      // DirectInput joystick state 
	BYTE * pbInput;
	DWORD dwInputSize;

	pPad->button = 0;
	pPad->stick_x = 0;
	pPad->stick_y = 0;

	s_InputSemaphore.Lock();

	for (i = 0; i < s_dwNumDevices; i++)
	{
		LPDIRECTINPUTDEVICE8 pDev;

		pDev = g_pDeviceInstances[i];
		// Skip any devices we couldn't create
		if (pDev == NULL)
			continue;


		hr = pDev->Poll(); 
		if (FAILED(hr))
		{
			// DInput is telling us that the input stream has been
			// interrupted. We aren't tracking any state between polls, so
			// we don't have any special reset that needs to be done. We
			// just re-acquire and try again.
			hr = pDev->Acquire();
			while ( hr == DIERR_INPUTLOST ) 
				hr = pDev->Acquire();

			// hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
			// may occur when the app is minimized or in the process of 
			// switching, so just try again later 
			continue; 
		}

		// Treat keyboard differently - use the diks structure
		if (i == 0)
		{
			pbInput = diks;
			dwInputSize = sizeof(diks);
		}
		else
		{
			pbInput = (BYTE*)&js;
			dwInputSize = sizeof(js);
		}

		ZeroMemory( pbInput, dwInputSize );
		hr = pDev->GetDeviceState( dwInputSize, pbInput );

		if (FAILED(hr)) 
			continue;		// The device should have already been acquired

		if (i == 0)
		{
			if (diks[DIK_ESCAPE] & 0x80)
				PostMessage(g_hMainWindow, WM_KEYDOWN, (WPARAM)VK_ESCAPE, 0);
			if (diks[DIK_F1] & 0x80)
				PostMessage(g_hMainWindow, WM_KEYDOWN, (WPARAM)VK_F1, 0);
			if (diks[DIK_F2] & 0x80)
				PostMessage(g_hMainWindow, WM_KEYDOWN, (WPARAM)VK_F2, 0);

		}


		for (c = 0; c < NUM_INPUT_IDS; c++)
		{
			// Only process this input if it is assigned to the device we're looking at
			if (g_CurrentButtons[c].dwDevice != i)
				continue;

			// Check if this input is a button or a stick
			if (!g_N64Buttons[c].bIsAxis)
			{
				// The n64 control is a button - we 
				// should only use the first offset
				if (pbInput[g_CurrentButtons[c].dwOfsA] & 0x80)
					pPad->button |= g_N64Buttons[c].button;

			}
			else
			{
				LONG val;
				// The n64 control is an axis. If bIsAxis
				// is set, then .dwOfsA refers to an axis value
				if (g_CurrentButtons[c].bIsAxis)
				{
					// Val should be in the range -80 to 80 already, 
					// As we set this property of the joystick during init
					val = ((LONG*)pbInput)[g_CurrentButtons[c].dwOfsA/4];
				}
				else
				{
					if (pbInput[g_CurrentButtons[c].dwOfsA] & 0x80)
						val = -ANALOGUE_STICK_RANGE;
					else if (pbInput[g_CurrentButtons[c].dwOfsB] & 0x80)
						val = +ANALOGUE_STICK_RANGE;
					else
						val = 0;
				}

				// Check which of the axis it was.
				// Later we might want the dpad and c buttons to be treated
				// as axis too
				if (c == INPUT_ID_STICK_X)
					pPad->stick_x = val;
				else if (c == INPUT_ID_STICK_Y)
					pPad->stick_y = -val;
			}
		}
	}
	s_InputSemaphore.Unlock();
    return S_OK;
}

